# $Id: Alignment_Module.pm,v 1.8 2007/03/13 04:25:16 cl Exp $

##
# Alignment_Controller represents the base class for all alignment controllers.
# It provides a default implementation of statistics generation and caching.
#


use strict;
use warnings;

use Bio::IScan::Global;
use Bio::IScan::Utilities;

use Bio::IScan::Server::Cache_Controller;


package Bio::IScan::Server::Alignment_Module;
	our @ISA = 'Bio::Root::Root';


sub Alignment_Module::New {
	my $root = Bio::Root::Root->new();
	$root->throw(-class => 'Bio::Root::BadParameter',
				 -text => "Undefined",
				 -value => "Aligment_Module::New");
}


sub new { # constructor
	my $class = shift;
	my $this = $class->SUPER::new(@_);
	
	my $parameters = ['iscan_instance'];
	my $optional_parameters = {};
	my $arguments = Utilities::Process_Arguments($parameters,
												 $optional_parameters,
												 @_);
	
	map { $this->{$_} = $arguments->{$_}; } keys %{ $arguments };
	
	return bless $this, $class;
}


sub align_irs { # instance
	my $this = shift;
	$this->throw(-class => 'Bio::Root::NotImplemented',
				 -text => "Aligment_Controller::align_irs",
				 -value => "Aligment_Controller::align_irs");		
}


sub align_drs { # instance
	my $this = shift;	
	$this->throw(-class => 'Bio::Root::NotImplemented',
				 -text => "Aligment_Controller::align_drs",
				 -value => "Aligment_Controller::align_drs");		
}


sub align_statistics { # instance
	my $this = shift;
	$this->throw(-class => 'Bio::Root::NotImplemented',
				 -text => "Aligment_Controller::align_statistics",
				 -value => "Aligment_Controller::align_statistics");		
}


sub align_sequences { # instance
	my $this = shift;
	$this->throw(-class => 'Bio::Root::NotImplemented',
				 -text => "Aligment_Controller::align_sequences",
				 -value => "Aligment_Controller::align_sequences");	
}


sub _new_alignment { # instance
	my $this = shift;
	$this->throw(-class => 'Bio::Root::NotImplemented',
				 -text => "Aligment_Controller::_new_alignment",
				 -value => "Aligment_Controller::_new_alignment");	
}


sub check_alignment_arguments { # instance
	my $this = shift;
	my $request = shift;
	
	if ((my $value = $request->ir_expand_out()) =~ /^[\d]+$/o) {
		if ($value < 0) {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "<ir_expand_out> must be an integer greater than 0\n" .
						 "\tValue: " . $value,
						 '-value' => 'Bio::Root::BadParameter');
		}
	}
	else {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "<ir_expand_out>" .
					 " must be an integer grater than 0\n" .
					 "\tValue: " . $value,
					 '-value' => 'Bio::Root::BadParameter');		
	}	
	
	if ((my $value = $request->ir_expand_in()) =~ /^[\d]+$/o) {
		if ($value < 0) {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "<ir_expand_in>" .
						 " must be an integer greater than 0\n" .
						 "\tValue: " . $value,
						 '-value' => 'Bio::Root::BadParameter');
		}
	}
	else {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "<ir_expand_in>" .
					 " must be an integer grater than 0\n" .
					 "\tValue: " . $value,
					 '-value' => 'Bio::Root::BadParameter');		
	}	

	if (defined (my $value = $request->ir_statistics_iterations())) {
		if ($value =~ /^[\d]+$/o) {
			if ($value < 0) {
				$this->throw('-class' => 'Bio::Root::BadParameter',
							 '-text' => "<ir_statistics_iterations> " .
							 "must be a positive integer or <undef>\n" .
							 "\tValue: " . $value,
							 '-value' => 'Bio::Root::BadParameter');
			}
		}
		else {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "<ir_statistics_iterations> " .
						 "must be a positive integer or <undef>\n" .
						 "\tValue: " . $value,
						 '-value' => 'Bio::Root::BadParameter');
		}
	}
	
	if ((my $value = $request->dr_expand_out()) =~ /^[\d]+$/o) {
		if ($value < 0) {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "<dr_expand_out>" .
						 " must be an integer greater than 0\n" .
						 "\tValue: " . $value,
						 '-value' => 'Bio::Root::BadParameter');
		}
	}
	else {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "<dr_expand_out>" .
					 " must be an integer grater than 0\n" .
					 "\tValue: " . $value,
					 '-value' => 'Bio::Root::BadParameter');		
	}	
	
	if ((my $value = $request->dr_expand_in()) =~ /^[\d]+$/o) {
		if ($value < 0) {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "<dr_expand_in>" .
						 " must be an integer greater than 0\n" .
						 "\tValue: " . $value,
						 '-value' => 'Bio::Root::BadParameter');
		}
	}
	else {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "<dr_expand_in>" .
					 " must be an integer grater than 0\n" .
					 "\tValue: " . $value,
					 '-value' => 'Bio::Root::BadParameter');		
	}
	
	
	if ((my $value = $request->alignment_match_value()) =~ /^[\d]+$/o) {
		if ($value < 0) {
			$this->throw('-class' => 'Bio::Root::BadParameter',
						 '-text' => "<alignment_match_value>" .
						 " must be an integer greater than 0\n" .
						 "\tValue: " . $value,
						 '-value' => 'Bio::Root::BadParameter');
		}
	}
	else {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "<alignment_match_value>" .
					 " must be an integer grater than 0\n" .
					 "\tValue: " . $value,
					 '-value' => 'Bio::Root::BadParameter');		
	}
}


sub _statistics_read_key { # instance
	my $this = shift;
	my $response = shift;
	my $sequence_id = shift;
	
	my $request = $response->search_request();
	
	my @key_elements = ();
	push(@key_elements, $sequence_id);	
	push(@key_elements, Global::Configuration_Parameter('alignment_type'));
	push(@key_elements, $request->ir_expand_out() + $request->ir_expand_in());
	push(@key_elements, $request->alignment_match_value());
	push(@key_elements, $request->alignment_mismatch_value());
	push(@key_elements, $request->alignment_gap_extension_value());
	push(@key_elements, $request->ir_statistics_iterations());
	
	return \@key_elements;
}


sub _statistics_write_key { # instance
	my $this = shift;
	my $response = shift;
	my $sequence_id = shift;
	
	my $request = $response->search_request();

	my @key_elements = ();
	push(@key_elements, $sequence_id);	
	push(@key_elements, Global::Configuration_Parameter('alignment_type'));
	push(@key_elements, $request->ir_expand_out() + $request->ir_expand_in());
	push(@key_elements, $request->alignment_match_value());
	push(@key_elements, $request->alignment_mismatch_value());
	push(@key_elements, $request->alignment_gap_extension_value());
	push(@key_elements, scalar @{ $response->ir_statistics($sequence_id) });
	
	return \@key_elements;
}


##
#
# The statistics database is organized by key where the key consists of the 
#	sequence ID, length of the sequences used for alignment, and number of
#	iterations for which scores were generated. The value associated with the 
#	key is a hash of all the scores produced for statistical purposes. For
#	example, the key 'NC_000000:50:10000' is for sequence NC_000000, using an
#	alignment length of 50 nucleotides 10,000 iterations.
#
sub generate_statistics { # instance
	my $this = shift;
	my $sequence_id = shift;
	my $response = shift;

	my $cache_name = $this->{'iscan_instance'}->db_name() . '.cache';
	my $cache_path = Global::Configuration_Parameter('cache_path');

	my $cache_controller = 
		Bio::IScan::Server::Cache_Controller->new(
			{ 'cache_path' => $cache_path,
			  'cache_name' => $cache_name });

	my $db_time = $this->{'iscan_instance'}->db_time();
	my $cache_file = $cache_controller->cache_file();
	
	my $statistics = undef;
	
	# If the cache file exists ('-e')
	if (-e $cache_file) {		
		# If any of the BLAST database files have been modified since the 
		#   statistics file was generated; the statistics file has to be
		#   regenerated	
		# Element '9' of 'stat' is the last modification time (mtime)
		if ($db_time > (stat($cache_file))[9]) {
			$this->_debug("Generating statistics: statistics too old\n");
			$statistics = $this->_generate_statistics($response, $sequence_id);
			$response->ir_statistics($sequence_id, $statistics);			

			my $cache_key = $cache_controller->generate_key(
								$this->_statistics_write_key($response,
															 $sequence_id));
			$cache_controller->cache_key($statistics, $cache_key);

			return $statistics;
		}
	}
	else {
		$this->_debug("Generating statistics: no statistics for database\n");		
		$statistics = $this->_generate_statistics($response, $sequence_id);
		$response->ir_statistics($sequence_id, $statistics);		
		
		my $cache_key = $cache_controller->generate_key(
							$this->_statistics_write_key($response,
														 $sequence_id));
		$cache_controller->cache_key($statistics, $cache_key);
		
		return $statistics;		
	}
	
	# The statistics file exists and the BLAST database has not been modified
	#   since the statistics file was generated, so the statistics file can be
	#   searched for the correct data.
	my $key_elements = $this->_statistics_read_key($response, $sequence_id);
	pop(@{ $key_elements });
	my $search_key = $cache_controller->generate_key($key_elements);
	$search_key .= '.*';
		
	# If the DB contains scores generated for more iterations than
	#   requested, the stored larger number of iterations are used.
	my @keys = $cache_controller->find_keys($search_key);
	my $max_iterations = 0;
	for my $key (@keys) {
		$this->_debug("Found statistics: ", $key);
		
		my @key_elements = split(/:/, $key);
		my $stored_iterations = pop(@key_elements);
		
		if ($stored_iterations > $max_iterations) {
			$max_iterations = $stored_iterations;
		}		
	}
	
	if ($max_iterations >= $response->search_request()->ir_statistics_iterations()) {
		push(@{ $key_elements }, $max_iterations);
		my $cache_key = $cache_controller->generate_key($key_elements);
		
		$this->_debug("Maximum statistics for iterations: " . $max_iterations . "\n");
		$response->search_request()->ir_statistics_iterations($max_iterations);
		
		my $statistics = $cache_controller->retrieve_key($cache_key);
		$this->_debug("Retrieved statistics: ", scalar @{ $statistics });
		
		$response->ir_statistics($sequence_id, $statistics);
		return $statistics;
	}	

	# If this point is reaced, the DB did not (a) contain scores generated for
	#   the requested length of sequence; or (b) contain scores generated for
	#   a number of iterations >= the requested number of iterations. Therefore,
	#   scores must be generated.
	$this->_debug("Generating statistics: sufficient iterations not found\n");	
	$statistics = $this->_generate_statistics($response, $sequence_id);
	$response->ir_statistics($sequence_id, $statistics);	

	my $cache_key = $cache_controller->generate_key(
						$this->_statistics_write_key($response,
													 $sequence_id));
	$cache_controller->cache_key($statistics, $cache_key);
	
	return $statistics;
}


##
#
sub _generate_statistics { # instance
	my $this = shift;
	
	my $response = shift;
	my $sequence_id = shift;

	my $sequence_length = $this->{'iscan_instance'}->db_sequence_length($sequence_id);
	
	my $sequence  = $this->{'iscan_instance'}->db_sequence($sequence_id);
	
	my $request = $response->search_request();
	my $length = $request->ir_expand_out() + $request->ir_expand_in();
	my $iterations = $request->ir_statistics_iterations();
	
	my @score_statistics = ();


	$this->{'iscan_instance'}->progress("Generating statistics for " . $sequence_id . "...\n");
	$this->{'iscan_instance'}->progress("%.....10......20......30......40......50......60......70......80......90.....100\n");
	my $progressed = 0;

	my $i = 0;
	my $statistics_failures = 0;
	while ($i < $iterations) {
		if ($this->{'iscan_instance'}->show_progress()) {
			my $percent = int(($i / $iterations) * 100);
			my $progress = int($percent / (100 / 80));
			if ($progress > $progressed) {
				$this->{'iscan_instance'}->progress("^" x ($progress - $progressed));
				$progressed = $progress;
			}
		}
		
		my $seq_string = undef;
		
		my $start_l = int(rand($sequence_length - $length));
		if ($start_l == 0) { $start_l = 1; }
		my $end_l = $start_l + $length;
		my $strand_l = int(rand(2));
		if ($strand_l == 2) {
			my $temp = $start_l;
			$start_l = $end_l;
			$end_l = $start_l;
		}
		$seq_string = $sequence->subseq($start_l, $end_l);
		# A Bio::Seq object cannot be created for a sequence consisting of
		#	nothing but "N"-s (the alphabet cannot be determined), so these
		#	sequences are thrown out.
		if ($seq_string =~ /^N*$/) { next; }
		my $sequence_l = Bio::Seq->new(-id => ($start_l . ".." . $end_l), 
									   -seq => $seq_string);

		my $start_r = $start_l;
		my $end_r = undef;
		while ($start_r == $start_l) {
			$start_r = int(rand($sequence_length - $length));
			if ($start_r == 0) { $start_r = 1; }
			$end_r = $start_r + $length;		
			my $strand_r = int(rand(2));
			if ($strand_l == 2) {
				my $temp = $start_r;
				$start_r = $end_r;
				$end_r = $start_r;				
			}			
		}
		$seq_string = $sequence->subseq($start_r, $end_r);
		# A Bio::Seq object cannot be created for a sequence consisting of
		#	nothing but "N"-s (the alphabet cannot be determined), so these
		#	sequences are thrown out.		
		if ($seq_string =~ /^N*$/) { next; }		
		my $sequence_r = Bio::Seq->new(-id => ($start_r . ".." . $end_r), 
									   -seq => $seq_string);
		
		eval {
			my $is_alignment = $this->align_statistics($sequence_l, $sequence_r, 
													   $response);

			$score_statistics[++$#score_statistics] = $is_alignment->score();			
		};
		# An exception is thrown when there is no alignment between the two
		#	sequences.
		# Unfortunantely, due to the lack of structure in Perl exceptions, this
		#	specific exception cannot be checked for. A warning is generated in case
		#	another type of exception is generated. Additionally, the exception can
		#	not continue, becuase it would be produced each time.	
		if ($@) {
			$this->_debug("Sequences can not be aligned\n" .
						  "\tSequence: " . $sequence_l->display_id() . "\n" .
						  "\t\t" . $sequence_l->seq() . "\n" .
						  "\tSequence: " . $sequence_r->display_id() . "\n" .
						  "\t\t" . $sequence_r->seq() . "\n" .
						  "\tException converted to warning:\n" .
						  $@);

			$statistics_failures += 1;
		}
		
		$i += 1;
	}
	
	$this->{'iscan_instance'}->progress("\n");
	
	if ($statistics_failures != 0) {
		$this->warn("Failure generating statistics\n" .
					"\tFailures: " . $statistics_failures);
	}
	
	$this->_debug("Generated statistics: ", scalar @score_statistics, 
				  " (", $i, ")");
	
	return \@score_statistics;
}


sub _debug { # instance
	my $this = shift;
	Global::Debug(__PACKAGE__, @_);
}


1;

__END__
