# $Id: CLUSTALW_Controller.pm,v 1.2 2007/03/13 04:21:09 cl Exp $

##
# This class is responsible for providing access to and control of many of the
#	features needed to configure, execute, and control CLUSTAL W.

use strict;
use warnings;

use Bio::Root::Root;
use Bio::AlignIO;
use Bio::SearchIO;
use Bio::SeqIO;

use File::Spec;

use Bio::IScan::Server::Controller;


package Bio::IScan::Server::CLUSTALW_Controller;
	our @ISA = 'Bio::IScan::Server::Controller';


##
# Create a new CLUSTALW_Controller
#
# @param path
#   Fully qualified path to the BLAST executable
# @param name
#   Database path to be used by BLAST

sub new {
	my $class = shift;
	my $this = $class->SUPER::new(@_);

	my $parameters = [ 'path', 'name' ];
	my $optional_parameters = {};

	my $arguments = $class->SUPER::_process_arguments($parameters, 
													  $optional_parameters, 
													  @_);

	map { $this->{$_} = $arguments->{$_}; } keys %{ $arguments };

	return bless $this, $class;
}


##
#

sub align_sequences { # instance
	my $this = shift;
	
	my $parameters = [ 'sequences' ];
	my $optional_parameters = { 'gap_value' => undef,
							    'extension_value' => undef
							  };

	my $arguments = Utilities::Process_Arguments($parameters, 
												 $optional_parameters, 
												 @_);

	# Temporary file for alignment
	my ($file_handle, $sequence_filename) = $this->temporary_file('clustalw-XXXX', '.fasta');
	my (undef, $alignment_filename) = $this->temporary_file('clustalw-XXXX', '.aln');
	my (undef, $tree_filename) = $this->temporary_file('clustalw-XXXX', '.dnd');
		
	for my $sequence (@{ $arguments->{'sequences'} }) {		
		print($file_handle ">", $sequence->display_id(), "\n", 
			  $sequence->seq(), "\n");
	}
	close $file_handle;
	
	# Execute DIALIGN
	my $command .= File::Spec->catfile($this->{path}, $this->{name}) . " ";
	$command .= "-infile=" . $sequence_filename . " ";
	$command .= "-align ";
	$command .= "-outorder=input ";
	$command .= "-outfile=" . $alignment_filename . " ";
	$command .= "-newtree=" . $tree_filename . " ";
	
	if (defined $arguments->{'gap_value'}) {
		$command .= "-gapopen=" . $arguments->{gap_value} . " ";
	}
	if (defined $arguments->{'extension_value'}) {
		$command .= "-gapext=" . $arguments->{extension_value} . " ";
	}

	$this->debug("Executing: ", $command, "\n");
	my $process = undef;
	my $status = open $process, $command . " |";	
	if ($status == 0) {
		$this->throw(-class => 'CLUSTALW_Controller_Exception',
					 -text => "Execution error: $command",
					 -value => "$! $?");
	}
	
	my $score = undef;
	while (<$process>) {
		# 1: score
		#                               1
		if ($_ =~ /Alignment\s+Score\s+([-\d]+)/io) {
			$score = $1;
		}
	}
	
	$status = close $process;
	if ($status == 0) {
		$this->throw(-class => 'CLUSTALW_Controller_Exception',
					 -text => "Error executing xdget: $command",
					 -value => "$! $?");
	}	
	
	my $align_io = Bio::AlignIO->new('-format' => 'clustalw', 
									 '-file' => $alignment_filename);
	
	my $alignment = $align_io->next_aln();
	$alignment->score($score);
	
	# The alignment coordinates have to be adjusted
	my @alignment_sequences = $alignment->each_seq();
	for (my $i = 0; $i < scalar @alignment_sequences; ++$i) {
		# If the supplied sequence was a <Bio::LocatableSeq>, recover the 
		#	coordinates; otherwise, the sequence does not have coordinates.
		if ($arguments->{'sequences'}->[$i]->isa('Bio::LocatableSeq')) {
			$alignment_sequences[$i]->start($arguments->{'sequences'}->[$i]->start());
			$alignment_sequences[$i]->end($arguments->{'sequences'}->[$i]->end());			
		}
	}

	return $alignment;
}


##
# Exception class for BLAST_Controller.
#
@CLUSTALW_Controller_Exception::ISA = 'Bio::Root::Exception';
$CLUSTALW_Controller_Exception::INVALID_SEQUENCE_TYPE = -1;


1;

