# $Id: dpalign_Alignment_Module.pm,v 1.9 2007/03/13 04:23:35 cl Exp $

##
# The <dpAlign_Alignment_Module> is an alignment module that performs a local
#	alignment of the left IR against the reverse complement of the right IR.
#


use strict;
# This module cannot use <warnings>, because each alignment module overrides the
#	<Alignment_Module::New> function defined in the subclass alignment module.

use Bio::IScan::Search_Request;
use Bio::IScan::Search_Response;
use Bio::IScan::Simple_Alignment;

use Bio::IScan::Utilities;

use Bio::Tools::dpAlign;


package Bio::IScan::Server::dpalign_Alignment_Module;
	our @ISA = 'Bio::IScan::Server::Alignment_Module';


sub Alignment_Module::New {
	return Bio::IScan::Server::dpalign_Alignment_Module->new(@_);
}


sub new { # constructor
	my $class = shift;
	my $this = $class->SUPER::new(@_);
	
	my $parameters = [];
	my $optional_parameters = {};
	my $arguments = Utilities::Process_Arguments($parameters,
												 $optional_parameters,
												 @_);
	
	map { $this->{$_} = $arguments->{$_}; } keys %{ $arguments };
	
	return bless $this, $class;
}


sub _new_alignment { # factory
	my $this = shift;
	return Bio::IScan::Simple_Alignment->new(@_);
}


sub check_alignment_arguments { # instance
	my $this = shift;
	my $request = shift;
	
	$this->SUPER::check_alignment_arguments($request);
	
	if ((my $sum = $request->ir_expand_out() + $request->ir_expand_in()) < 3) {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "The sum of <ir_expand_in> and <ir_expand_out> " . 
					 "must be grater than 3\n" .
					 "\tSum: " . $sum,
					 '-value' => 'Bio::Root::BadParameter');		
	}
	
	if ((my $sum = $request->dr_expand_out() + $request->dr_expand_in()) < 3) {
		$this->throw('-class' => 'Bio::Root::BadParameter',
					 '-text' => "The sum of <dr_expand_in> and <dr_expand_out> must be grater than 3\n" .
					 "\tSum: " . $sum,
					 '-value' => 'Bio::Root::BadParameter');		
	}
	
}


sub align_irs { # framework
	my $this = shift;
	
	my $response = shift;
	my $search_sequence_l = shift;
	my $search_sequence_r = shift;
	my $sequence_id = shift;

	my $request = $response->search_request();

	my ($align_sequence_l, $align_sequence_r) = undef;
	
	# Reverse complement the right (downstream) sequence.
	$align_sequence_l = $search_sequence_l;
	$align_sequence_r = $search_sequence_r->revcom();

	$this->_debug("L. IR: ", 
				  $align_sequence_l->start(), "..", $align_sequence_l->end());
	$this->_debug("R. IR: ", 
				  $align_sequence_r->start(), "..", $align_sequence_r->end());
	
	my $alignment = $this->align_sequences($align_sequence_l, 
										   $align_sequence_r,
										   $response);

	# The start and end coordinates of the IRs have to be adjusted to genomic
	#   coordinates, because (counter to the BioPerl documentation) the
	#   alignment does not keep track of them.
	my $ir_l = $alignment->get_seq_by_pos(1);
	my $ir_r = $alignment->get_seq_by_pos(2);
	
	$this->_debug("L. IR Aln.: ", 
				  $ir_l->start(), "..", $ir_l->end(), "\t", $ir_l->seq());
	$this->_debug("R. IR Aln.: ", 
				  $ir_r->start(), "..", $ir_r->end(), "\t", $ir_r->seq());
	
	if ($align_sequence_l->strand() < 0) {
		my $ir_l_start = $ir_l->start();
		$ir_l->start($align_sequence_l->end() - $ir_l->end() + 1);
		$ir_l->end($align_sequence_l->end() - $ir_l_start + 1);
		
		$ir_r->start($align_sequence_r->start() + $ir_r->start() - 1);
		$ir_r->end($align_sequence_r->start() + $ir_r->end() - 1);
	}
	else {
		$ir_l->start($align_sequence_l->start() + $ir_l->start() - 1);
		$ir_l->end($align_sequence_l->start() + $ir_l->end() - 1);
		
		my $ir_r_start = $ir_r->start();
		$ir_r->start($align_sequence_r->end() - $ir_r->end() + 1);
		$ir_r->end($align_sequence_r->end() - $ir_r_start + 1);
	}
	
	$this->_debug("L. IR Aln. (Adj.): ", $ir_l->start(), "..", $ir_l->end());
	$this->_debug("R. IR Aln. (Adj.): ", $ir_r->start(), "..", $ir_r->end());
	
	my $is_alignment = $this->_new_alignment(
						{ 'simple_align' => $alignment,
						  'match_score' => $request->alignment_match_value(), 
						  'mismatch_score' => $request->alignment_mismatch_value(),
						  'gap_score' => abs($request->alignment_gap_extension_value()), 
						  'ext_score' => abs($request->alignment_gap_extension_value()) 
						});
	
	# Calculate alignment statistics if necessary
	if ($request->ir_generate_statistics()) {
		my $statistics = $response->ir_statistics($sequence_id);

		my $percentile = $is_alignment->calculate_percentile($statistics);
		$is_alignment->percentile($percentile);
	}
	
	return ($is_alignment, $is_alignment);
}


sub align_drs { # framework
	my $this = shift;
	
	my $response = shift;	
	my $search_sequence_l = shift;
	my $search_sequence_r = shift;

	my $request = $response->search_request();
	
	my $alignment = $this->align_sequences($search_sequence_l, 
										   $search_sequence_r, 
										   $response);
	
	# The start and end coordinates of the IRs have to be adjusted to genomic
	#   coordinates, because (counter to the BioPerl documentation) the
	#   alignment does not keep track of them.
	my $dr_l = $alignment->get_seq_by_pos(1);
	my $dr_r = $alignment->get_seq_by_pos(2);
	
	$this->_debug("L. DR Aln.: ", 
				  $dr_l->start(), " -> ", $dr_l->end(), "\t", $dr_l->seq());
	$this->_debug("R. DR Aln.: ", 
				  $dr_r->start(), " -> ", $dr_r->end(), "\t", $dr_r->seq());
	
	# The start and end coordinates of the DRs have to be adjusted to genomic
	#   coordinates, because (counter to the BioPerl documentation) the
	#   alignment does not keep track of them.
	if ($search_sequence_l->strand() < 0) {
		my $dr_l_start = $dr_l->start();
		$dr_l->start($search_sequence_l->end() - $dr_l->end() + 1);
		$dr_l->end($search_sequence_l->end() - $dr_l_start + 1);

		my $dr_r_start = $dr_r->start();
		$dr_r->start($search_sequence_r->end() - $dr_r->end() + 1);
		$dr_r->end($search_sequence_r->end() - $dr_r_start + 1);
	}
	else {
		$dr_l->start($search_sequence_l->start() + $dr_l->start() - 1);
		$dr_l->end($search_sequence_l->start() + $dr_l->end() - 1);

		$dr_r->start($search_sequence_r->start() + $dr_r->start() - 1);
		$dr_r->end($search_sequence_r->start() + $dr_r->end() - 1);
	}
	
	$this->_debug("L. DR Aln. (Adj.): ", 
				  $dr_l->start(), " -> ", $dr_l->end(), "\t", $dr_l->seq());
	$this->_debug("R. DR Aln. (Adj.): ", 
				  $dr_r->start(), " -> ", $dr_r->end(), "\t", $dr_r->seq());
	
	my $is_alignment = $this->_new_alignment(
						{ simple_align => $alignment,
						  match_score => $request->alignment_match_value(), 
						  mismatch_score => $request->alignment_mismatch_value(),
						  gap_score => abs $request->alignment_gap_extension_value(), 
						  ext_score => abs $request->alignment_gap_extension_value() 
						});
	
	return $is_alignment;	
}


sub align_statistics { # framework
	my $this = shift;
	my $sequence_1 = shift;
	my $sequence_2 = shift;
	my $response = shift;
	
	my $is_alignment = $this->align_sequences($sequence_1, $sequence_2, $response);

	return $is_alignment;
}


sub align_sequences { # framework
	my $this = shift;
	my $sequence_1 = shift;
	my $sequence_2 = shift;
	my $response = shift;
	
	my $request = $response->search_request();

	my $align_factory = 
		Bio::Tools::dpAlign->new('-match' => abs($request->alignment_match_value()), 
								 '-mismatch' => -abs($request->alignment_mismatch_value()),
								 '-gap' => abs($request->alignment_gap_extension_value()), 
								 '-ext' => abs($request->alignment_gap_extension_value()),
								 '-alg' => $Bio::Tools::dpAlign::DPALIGN_LOCAL_MILLER_MYERS);
	
	my $alignment = $align_factory->pairwise_alignment($sequence_1, $sequence_2);
	
	my $is_alignment = $this->_new_alignment(
						{ 'simple_align' => $alignment,
						  'match_score' => abs($request->alignment_match_value()),
						  'mismatch_score' => -abs($request->alignment_mismatch_value()),
						  'gap_score' => -abs($request->alignment_gap_extension_value()), 
						  'ext_score' => -abs($request->alignment_gap_extension_value())
						});	
	
	
	return $is_alignment;
}


sub _debug { # instance
	my $this = shift;
	Global::Debug(__PACKAGE__, @_);
}


1;

__END__
