# $Id: Simple_Alignment.pm,v 1.7 2007/03/13 04:37:55 cl Exp $

##
# A 'Simple_Alignment' is just the alignment of one sequence against another. A
#	'Simple_Alignment' is used to store alignments when there is no template
#	sequence being aligned.
#


use strict;
use warnings;

use Bio::SimpleAlign;
use Bio::Root::Root;

use Carp;


package Bio::IScan::Simple_Alignment;
	our @ISA = 'Bio::SimpleAlign';


sub new { # constructor
	my $class = shift;
	
	my ($args) = @_;
	# Convert arguments to lowercase
	$args->{map(lc($_), keys %{$args})} = values %{$args};
	
	for my $test ('simple_align') {
		if (!exists $args->{$test}) {
			$class->throw(-class => 'Bio::Root::BadParameter',
						  -text => "Argument must be specified: $test",
						  -value => $test);
		}
	}
	
	my $copy = $args->{simple_align};
	my $this = $class->SUPER::new(-source => $copy->source(), 
								  -verbose => $copy->verbose());
	
	for my $sequence ($copy->each_seq()) {
		$this->add_seq($sequence);
	}
	
	$this->{'wrapped_alignment'} = $copy;

	for my $argument ('match_score',
					  'mismatch_score',
					  'gap_score',
					  'ext_score') {
		if (exists $args->{$argument}) {
			$this->{$argument} = $args->{$argument};
		}
	}
	
	# Since it can be done here, sequence the calls correctly.
	$this->{matches} = $this->_calculate_matches($args->{simple_align});
	$this->{gaps} = $this->_calculate_gaps($args->{simple_align});	
	$this->{mismatches} = $this->_calculate_mismatches($args->{simple_align});
	
	$this->score($this->_calculate_score($args->{simple_align}));
	
	$this->{'percentile'} = 0;
	
	return bless $this, $class;		
}


sub wrapped_alignment { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Alignment are immutable.",
					 -value => @_);
	}
	
	return $this->{'wrapped_alignment'};
}


sub left_sequence { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Alignment are immutable.",
					 -value => @_);
	}
	
	return $this->get_seq_by_pos(1);
}


sub right_sequence { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Alignment are immutable.",
					 -value => @_);
	}
	
	return $this->get_seq_by_pos(2);
}


sub matches { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Alignment are immutable.",
					 -value => @_);
	}
	
	return $this->{matches};
}


sub mismatches { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Alignment are immutable.",
					 -value => @_);
	}
	
	return $this->{mismatches};	
}


sub gaps { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Alignment are immutable.",
					 -value => @_);
	}
	
	return $this->{gaps};		
}


sub percentile { # instance
	my $this = shift;
	
	if (@_) {
		$this->{percentile} = shift;
	}
	
	return $this->{percentile};
}


sub p_value { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Instances of Alignment are immutable.",
					 -value => @_);
	}
	
	return 1 - ($this->percentile() / 100);
}


sub _calculate_score { # instance
	my $this = shift;
	
	my $alignment = shift;

	for my $test ('match_score',
				  'mismatch_score',
				  'gap_score',
				  'ext_score') {
		if (!exists $this->{$test}) {
			$this->throw(-class => 'Alignment_Exception',
						 -text => "A score cannot be calculated because the following value was not specified: $test",
						 -value => $test);
		}
	}

	if (!exists $this->{matches}) {
		$this->{matches} = _calculate_matches($alignment);
	}
	
	if (!exists $this->{gaps}) {
		$this->{gaps}= _calculate_gaps($alignment);
	}

	if (!exists $this->{mismatches}) {
		$this->{mismatches} = _calculate_mismatches($alignment);
	}	
	
	my $score = ($this->matches() * $this->{match_score}) +
				($this->mismatches() * $this->{mismatch_score}) +
				($this->gaps() * $this->{gap_score});
	
	return $score;
}


sub _calculate_matches { # instance
	my $this = shift;
	
	my $alignment = shift;

	my $matches = 0;
	my $line = $alignment->match_line();
	if (defined $line) {
		while ($line =~ /\*/g) { ++$matches; }
	}

	return $matches;
}


sub _calculate_mismatches { # instance
	my $this = shift;
	
	my $alignment = shift;

	if (!exists $this->{gaps}) {
		$this->{gaps} = _calculate_gaps($alignment);
	}

	my $mismatches = 0;
	my $line = $alignment->consensus_string(100);
	if (defined $line) {
		while ($line =~ /\?/g) {++$mismatches; }
	}
	$mismatches -= $this->{gaps};
	
	return $mismatches;
}


sub _calculate_gaps { # instance
	my $this = shift;
	
	my $alignment = shift;

	my $gaps = 0;
	
	my $gap_char = '-';
	for my $sequence ($this->each_seq()) {
		my $line = $sequence->seq();
		if (defined $line) {
			while ($line =~ /\-/g) { ++$gaps; }
		}
	}

	return $gaps;
}


sub calculate_percentile { # instance
	my $this = shift;
	my $statistics = shift;
	
	if (!defined $statistics || scalar @{ $statistics } < 1) {
		return 0;
	}

	my @keys = sort({ $a <=> $b } @{ $statistics });
	my $lower_scores = 0;
		
	for my $key (@keys) {
		if ($this->score() >= $key) { $lower_scores += 1; }
		else { last; }
	}

	my $total_scores = scalar @keys;
	
	return ($lower_scores == 0) ? 0 : ($lower_scores / $total_scores) * 100;
}

	
##
# Exception class for BLAST_Controller.
#
@Alignment_Exception::ISA = 'Bio::Root::Exception';

1;
