#!/usr/bin/perl

# $Id: test.t,v 1.1 2006/12/27 05:56:02 cl Exp $


use strict;
use warnings;


use Test::More 'no_plan';

use lib '.';

use Bio::IScan::Global;
use Bio::IScan::Search_Request;

use Bio::IScan::Server::IScan;

use Bio::SeqIO;

use File::Spec;


my $db_path = "tutorial";
my $db_name = "Escherichia_coli_K12";

my $query_file = "tutorial/IS5.faa";


# Load configuration information for 'IScan::Configuration' and 
#	'IScan::Search::Configuration'
require iscan_config;

# '$IScan::Configuration' comes from 'iscan_config.pm'
my $configuration = $IScan::Configuration;

ok(defined $configuration, "Reading configuration");

# Store all of the configuration parameters for global access
for my $key (keys %{ $IScan::Configuration }) {
	Global::Configuration_Parameter($key, $configuration->{$key});
}


$configuration->{'db_path'} = $db_path;
$configuration->{'db_name'} = $db_name;

my $iscan = Bio::IScan::Server::IScan->new($configuration);

ok(defined $iscan, "Creating IScan instance");

my $search_request = 
	Bio::IScan::Search_Request->new($IScan::Search::Configuration);

ok(defined $search_request, "Creating search request");


# The IS name is based on the file name.
my (undef, undef, $query_filename) = File::Spec->splitpath($query_file);
my ($query_name) = ($query_filename =~ /^(.*)\..*$/o);

$search_request->search_name($query_name);

my @queries = ();    
my $fasta_io = Bio::SeqIO->new(-format => 'fasta', -file => $query_file);
while ( (my $sequence = $fasta_io->next_seq()) ) {
	if ($sequence->display_id() eq "IR-LT") {
		$search_request->ir_upstream_profile($sequence);
	}
	elsif ($sequence->display_id() eq "IR-RT") {
		$search_request->ir_downstream_profile($sequence);
	}
	else {
		$queries[++$#queries] = $sequence;
	}
}
$search_request->queries(\@queries);

ok(scalar @queries > 0, "Reading querie(s)");

my %search_arguments = (%{ $IScan::Search::Configuration });
$search_arguments{'queries'} = \@queries;

my $search_response = $iscan->search($search_request); 

ok(defined $search_response, "Searching");

my $results = $search_response->insertion_sequences();

my $blast_controller = Bio::IScan::Server::BLAST_Controller->new(
						  { 'blast_path' => $configuration->{'blast_path'},
							  'db_name' => $db_name,
							  'db_path' => $db_path,
							  'filter_path' => $configuration->{'blast_path'},
							  'matrix_path' => $configuration->{'blast_path'},
							  'db_genetic_code' => $configuration->{'blast_path'},
						  });


for my $is (@{ $results }) {
	diag("Verifying IS " . $is->display_id());
	
	# IS sequence
	my $sequence = $is->is_seq();
	my $sequence_string = $sequence->seq();
	
	my $start = $is->is_seq()->start();
	my $end = $is->is_seq()->end();
	my $test_string = 
		$blast_controller->retrieve_sequence($is->source_locus(), $start, $end);
	my $test_sequence = Bio::LocatableSeq->new('-seq' => $test_string,
											   '-start' => $start,
											   '-end' => $end);
	if ($is->strand() < 0) {
		$test_sequence = $test_sequence->revcom();
	}
	
	is($sequence_string, $test_sequence->seq(), "Verifying IS sequence");
	is($sequence->start(), $is->start());
	is($sequence->end(), $is->end());
	if ($is->strand() < 0) {
		is($is->left_ir_alignment()->left_sequence()->end(), $is->end());
		is($is->right_ir_alignment()->right_sequence()->start(), $is->start());
	}
	else {
		is($is->left_ir_alignment()->left_sequence()->start(), $is->start());
		is($is->right_ir_alignment()->right_sequence()->end(), $is->end())
	}
	

	# Left IR
	$sequence = $is->left_ir_alignment_seq();
	$sequence_string = $sequence->seq();
	$sequence_string =~ y/-//d;

	$start = $is->left_ir_alignment()->left_sequence()->start();
	$end = $is->left_ir_alignment()->left_sequence()->end();
	$test_string = 
		$blast_controller->retrieve_sequence($is->source_locus(), $start, $end);
	$test_sequence = Bio::LocatableSeq->new('-seq' => $test_string,
											'-start' => $start,
											'-end' => $end);
	if ($is->strand() < 0) {
		$test_sequence = $test_sequence->revcom();
	}
	
	is($sequence_string, $test_sequence->seq(), "Verifying left IR");
	is($is->left_ir_alignment()->left_sequence()->start(), $sequence->start());
	is($is->left_ir_alignment()->left_sequence()->end(), $sequence->end());	
	
	
	# Right IR
	$sequence = $is->right_ir_alignment_seq();
	$sequence_string = $sequence->seq();
	$sequence_string =~ y/-//d;
		
	$start = $is->right_ir_alignment()->right_sequence()->start();
	$end = $is->right_ir_alignment()->right_sequence()->end();
	$test_string = 
		$blast_controller->retrieve_sequence($is->source_locus(), $start, $end);
	$test_sequence = Bio::LocatableSeq->new('-seq' => $test_string,
											'-start' => $start,
											'-end' => $end);
	if ($is->strand() > 0) {
		$test_sequence = $test_sequence->revcom();
	}
	
	is($sequence_string, $test_sequence->seq(), "Verifying right IR");
	is($is->right_ir_alignment()->right_sequence()->start(), $sequence->start());
	is($is->right_ir_alignment()->right_sequence()->end(), $sequence->end());	
	
	
	$sequence = $is->left_dr_alignment_seq();
	$sequence_string = $sequence->seq();
	$sequence_string =~ y/-//d;
		
	$start = $is->dr_alignment()->left_sequence()->start();
	$end = $is->dr_alignment()->left_sequence()->end();
	$test_string = 
		$blast_controller->retrieve_sequence($is->source_locus(), $start, $end);
	$test_sequence = Bio::LocatableSeq->new('-seq' => $test_string,
											'-start' => $start,
											'-end' => $end);
	if ($is->strand() < 0) {
		$test_sequence = $test_sequence->revcom();
	}
	
	is($sequence_string, $test_sequence->seq(), "Verifying left DR");

	
	$sequence = $is->right_dr_alignment_seq();
	$sequence_string = $sequence->seq();
	$sequence_string =~ y/-//d;
		
	$start = $is->dr_alignment()->right_sequence()->start();
	$end = $is->dr_alignment()->right_sequence()->end();
	$test_string = 
		$blast_controller->retrieve_sequence($is->source_locus(), $start, $end);
	$test_sequence = Bio::LocatableSeq->new('-seq' => $test_string,
											'-start' => $start,
											'-end' => $end);
	if ($is->strand() < 0) {
		$test_sequence = $test_sequence->revcom();
	}
	
	is($sequence_string, $test_sequence->seq(), "Verifying right DR");
}
	
