# $Id: BLAST_Controller.pm,v 1.10 2006/11/29 07:03:09 cl Exp $

##
# This class is responsible for providing access to and control of many of the
#   features needed to configure, execute, and control BLAST.
#
# This class is intended to manipulate the Washington University version of the
#   BLAST application (WU BLAST). There should be no WU BLAST specific code---but
#   it hasn't been tested against another version of BLAST. This class can be
#   subclassed to change the functionality of any code. An attempt has been made
#   to construct the class in such a way that it can be subclassed and individual
#   methods overriden in a framework style if changes need to be made.
#


use strict;

use Bio::IScan::Server::Controller;

use Bio::Root::Root;
use Bio::SearchIO;
use Bio::SeqIO;

use File::Spec;


package Bio::IScan::Server::BLAST_Controller;
	our @ISA = 'Bio::IScan::Server::Controller';


##
# Create a new BLAST_Controller
#
# @param -blast
#   Fully qualified path to the BLAST executable
# @param -db_path
#   Database path to be used by BLAST
# @param -filter_path
#   Filter path to be used by BLAST
# @param -matrix_path
#   Matrix path to be used by BLAST
#
sub new {
	my $class = shift;

	my $parameters = ['blast_path', 
					  'db_path',
					  'db_name',
					  'filter_path', 
					  'matrix_path'];
	my $optional_parameters = ['blast_db_genetic_code'];
	
	my $arguments = $class->SUPER::_process_arguments($parameters, 
													  $optional_parameters, 
													  @_);

	if (defined $arguments->{'blast_keep_files'}) {
		$arguments->{'keep_files'} = $arguments->{'blast_keep_files'};
	}
	
	if (defined $arguments->{'blast_temporary_path'}) {
		$arguments->{'temporary_path'} = $arguments->{'blast_temporary_path'};
	}
	
	my $this = $class->SUPER::new($arguments);

	map { $this->{$_} = $arguments->{$_}; } keys %{ $arguments };
	
	if (!defined $this->{'db_genetic_code'}) {
		$this->{'db_genetic_code'} = 1;
	}
	
	return bless $this, $class;
}


##
#
sub db_filename { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "BLAST_Controller is immutable: the database filename cannot be altered",
					 -value => @_);
	}
	
	return $this->{db_name};
}


##
# Returns an array of the files used for the BLAST database this controller is
#   associated with.
#
# NOTE: This method is specific to 'xdformat'-ed databases.
#
# @return @strings
#   Array containing the filenames, with whatever path was specified in the
#		constructor (see @{new}), of the files associated with this BLAST
#		database.
#
sub db_files { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "BLAST_Controller is immutable: the database files cannot be altered",
					 -value => @_);
	}
	
	my @filenames = ();
	
	$filenames[++$#filenames] = 
		File::Spec->catfile($this->{db_path}, $this->{db_name} . '.xnd');
	$filenames[++$#filenames] = 
		File::Spec->catfile($this->{db_path}, $this->{db_name} . '.xni');
	$filenames[++$#filenames] = 
		File::Spec->catfile($this->{db_path}, $this->{db_name} . '.xns');
	$filenames[++$#filenames] = 
		File::Spec->catfile($this->{db_path}, $this->{db_name} . '.xns');
	
	return @filenames;
}


##
#
sub db_length { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "BLAST_Controller is immutable: the database length cannot be altered",
					 -value => @_);
	}

	my $command = File::Spec->catfile($this->{blast_path}, 'xdformat');
	$command .= " -n ";
	$command .= " -i ";
	$command .= " " . File::Spec->catfile($this->{db_path}, $this->{db_name});
	
	$this->debug("Executing: ", $command, "\n");
	my $process = undef;
	my $status = open $process, $command . " 2>" . File::Spec->devnull() . "|";
	if ($status == 0) {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Error executing command: $command",
					 -value => "$! $?");
	}

	my $db_length = undef;
	while (<$process>) {
		# 1: Database size
		#								               1
		if ($_ =~ /sequences\s+\(letters\):\s+[\d,]+\s+\(([\d\,]+)\)/o) {
			$db_length = $1;
			# Remove commas (",")
			$db_length =~ tr/\,//d;
		}
	}
	
	$status = close $process;
	if ($status == 0) {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Error executing xdget: $command",
					 -value => "$! $?");
	}	
	
	if (!defined $db_length) {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "BLAST database size could not be found");
	}
	
	return $db_length;
}


##
#
sub db_sequences { # instance
	my $this = shift;
	
	if (@_) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "BLAST_Controller is immutable",
					 -value => @_);
	}

	my $command = File::Spec->catfile($this->{blast_path}, 'xdformat');
	$command .= " -n ";
	$command .= " -r ";
	$command .= " -m 2 ";
	$command .= " " . File::Spec->catfile($this->{db_path}, $this->{db_name});
	
	$this->debug("Executing: ", $command, "\n");
	my $process = undef;
	my $status = open $process, $command . " |";
	if ($status == 0) {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Error executing command: $command",
					 -value => "$! $?");
	}
	
	my @db_sequences = ();
	while (<$process>) {
		my ($sequence, undef) = split /[[:space:]]/, $_;
		push @db_sequences, $sequence;
	}
	
	$status = close $process;
	if ($status == 0) {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Error executing command: $command",
					 -value => "$! $?");
	}	
		
	return @db_sequences;
}


##
#
# @param identifier
# @param value
#
# @return
#
# @throw Bio::Root::BadParameter
#
#
# @throw BLAST_Controller_Exception
#
#
sub sequence_length { # instance
	my $this = shift;
	
	my @parameters = ('identifier');

	my ($args) = @_;
	# Convert arguments to lowercase
	%{ $args } = map { lc($_) => $args->{$_} } keys %{ $args };

	# Make sure all the arguments are valid
	for my $argument (keys %{ $args }) {
		my $found = 0;
		for my $parameter (@parameters) {
			if ($argument eq $parameter) {
				$found = 1;
			}
		}

		if (!$found) {
			$this->throw(-class => 'Bio::Root::BadParameter',
						 -text => "Invalid argument: $argument",
						 -value => $argument);
		}
	}

	# The below can be replaced with an 'any' in Perl 6.
	for my $test ('identifier') {
		if (!exists $args->{$test}) {
			$this->throw(-class => 'Bio::Root::BadParameter',
						 -text => "Argument must be specified: $test",
						 -value => $test);
		}
	}
	
	my $value = $args->{value};
	
	my $command = File::Spec->catfile($this->{blast_path}, 'xdget');
	$command .= " -n ";
	$command .= " -m 2 ";
	$command .= " " . File::Spec->catfile($this->{db_path}, $this->{db_name});
	$command .= " " . $args->{'identifier'};
	
	$this->debug("Executing: ", $command, "\n");
	my $process = undef;
	my $status = open $process, $command . " |";
	if ($status == 0) {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Error executing xdget: $command",
					 -value => "$! $?");
	}

	my $sequence_length = undef;
	while (<$process>) {
		# 'xdget' reports "Not fond:\t<identifier>"
		if ($_ =~ /Not found/io) {
			$this->throw(-class => 'BLAST_Controller_Exception',
						 -text => "Sequence size could not be found:\n" .
						 "\tIdentifier: " . $args->{'identifier'} . "\n");
		}
		
		# 'xdget' reports "<tag>|<identifier>[|<identifier>]\t<length>"
		#   E.g.: "gb|<accession>|<locus>\t<length>
		if ($_ =~ /.*$args->{'identifier'}.*[[:space:]]+([[:digit:]]+)$/) {
			$sequence_length = $1;
		}
	}

	$status = close $process;
	if ($status == 0) {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Error executing xdget: $command",
					 -value => "$! $?");
	}

	return $sequence_length;
}


##
# Run BLAST against an AA sequence.
#
# @param -query
#   Query sequence to blast 
#
# @return
#   
sub blast { # instance
	my $this = shift;
	
	my $query = shift;
	
	my ($in_fh, $in_file_name) = $this->temporary_file('blast-XXXX', '.fasta');
	my (undef, $out_file_name) = $this->temporary_file('blast-XXXX', '.blast');
	
	print $in_fh ">", $query->display_id(), $query->description(), "\n";
	print $in_fh $query->seq();
	
	my $query_type = $query->alphabet();
	my $blast_command = undef;
	if ($query_type eq 'dna') {
		$blast_command = 'blastn';
	}
	elsif ($query_type eq 'protein') {
		$blast_command = 'blastp';
	}
	elsif ($query_type eq 'rna') {
		$this->warn("BLAST-ing an RNA sequence against a nucleotide database");
		$blast_command = 'blastn';
	}
	else {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Cannot execute BLAST for an unknown sequence type: $query_type",
					 -value => $query_type);
	}
	
	my $blast_command = $this->get_blast_command($blast_command, 
												 $in_file_name, 
												 $out_file_name);
	
	my $value = system("$blast_command");
	if ($value != 0) {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Error executing BLAST ($blast_command): $!",
					 -value => $!);
	}
	
	my $blast_io = Bio::SearchIO->new(-format => 'blast', 
									  -file => $out_file_name);
	
	return $blast_io;
}


##
# Run BLAST against an AA sequence.
#
# @param -query
#   Query sequence to blast 
#
# @return
#   
sub blast_translate { # instance
	my $this = shift;
	
	my $query = shift;
	
	my ($in_fh, $in_file_name) = $this->temporary_file('blast-XXXX', '.fasta');
	my (undef, $out_file_name) = $this->temporary_file('blast-XXXX', '.blast');

	print $in_fh ">", $query->display_id(), $query->description(), "\n";
	print $in_fh $query->seq();
		
	my $query_type = $query->alphabet();
	my $blast_command = undef;
	if ($query_type eq 'dna') {
		$blast_command = 'tblastx';
	}
	elsif ($query_type eq 'protein') {
		$blast_command = 'tblastn';
	}
	elsif ($query_type eq 'rna') {
		$this->warn("BLAST-ing an RNA sequence against a nucleotide database");
		$blast_command = 'tblastx';
	}
	else {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Cannot execute BLAST for an unknown sequence type: $query_type",
					 -value => $query_type);
	}
			
	my $blast_command = $this->get_blast_command($blast_command, 
												 $in_file_name, 
												 $out_file_name);
	$blast_command .= " -dbgcode=" . $this->{db_genetic_code};	
	
	my $value = system("$blast_command");
	if ($value != 0) {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Error executing BLAST ($blast_command): $!",
					 -value => $!);
	}
	
	my $blast_io = Bio::SearchIO->new(-format => 'blast', 
									  -file => $out_file_name);
	
	return $blast_io;
}


##
# Run BLAST against an AA sequence.
#
# @param -query
#   Query sequence to blast 
#
# @return
#   
sub blast_AA_nucleotide { # instance
	my $this = shift;
	
	my $query = shift;

	my ($in_fh, $in_file_name) = temporary_file('blast-XXXX', '.fasta');
	my (undef, $out_file_name) = temporary_file('blast-XXXX', '.blast');
		
	print $in_fh ">" . $in_file_name . "\n";
	print $in_fh $query;

	my $blast_command = $this->get_blast_command('tblastn', 
												 $in_file_name, 
												 $out_file_name);
	$blast_command .= " -dbgcode=" . $this->{db_genetic_code};
	

	my $value = system("$blast_command");
	if ($value != 0) {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Error executing BLAST ($blast_command): $!",
					 -value => $!);
	}
	
	my $blast_io = Bio::SearchIO->new(-format => 'blast', 
									  -file => $out_file_name);
	
	return $blast_io;
}


sub get_blast_command { # instance
	my $this = shift;
	
	my $type = shift;
	my $input_file = shift;
	my $output_file = shift;
	if (!defined($type)) {
		$this->throw(-class => 'Bio::Root::BadParameter',
					 -text => "Argument must be specified",
					 -value => "type");		
	}
	
	my $command = undef;
	$command = File::Spec->catfile($this->{blast_path}, $type);
	
	$command .= " " . File::Spec->catfile($this->{db_path}, $this->{db_name});
	$command .= " " . $input_file;	
	$command .= " -o " . $output_file;
	
	# The following messages are visible to the user and would probably only
	#   be confusing.
	$command .= " -notes -warnings -errors";
	
	return $command;
}


sub retrieve_seq { # instance
	my $this = shift;
	
	my $id = shift;
	my $start = shift;
	my $end = shift;


	my $command = File::Spec->catfile($this->{blast_path}, 'xdget');
	$command .= " -n ";
	$command .= " -a " . $start;
	$command .= " -b " . $end;
	$command .= " " . File::Spec->catfile($this->{db_path}, $this->{db_name});
	$command .= " " . $id;
	my $process = undef;
	my $status = open $process, $command . " |";
	if ($status == 0) {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Error executing xdget: $command",
					 -value => "$! $?");
	}
	
	my $fasta_io = Bio::SeqIO->new(-format => 'fasta', -fh => $process);
	
	my $sequence = $fasta_io->next_seq();
	
	$status = close $process;
	if ($status == 0) {
		$this->throw(-class => 'BLAST_Controller_Exception',
					 -text => "Error executing xdget: $command",
					 -value => "$! $?");
	}
	
	return $sequence;
}


sub retrieve_sequence { # instance
	my $this = shift;
	
	my $sequence = $this->retrieve_seq(@_);
	
	return $sequence->seq();
}


##
# Exception class for BLAST_Controller.
#
@BLAST_Controller_Exception::ISA = 'Bio::Root::Exception';

1;
